const assert = require('assert');


class Reporter {

  constructor(opts = {}) {
    this.flows       = {};
    this.tasks       = [];
    this.tasksPassed = [];
    this.tasksFailed = [];
    this.flowsPassed = [];
    this.flowsFailed = [];
    this.startedAt   = null;
    this.finishedAt  = null;
    this.taskTime    = null;
    this.flowTime    = null;
    this.totalTime   = null;
    this.isPass      = false;

    const noop       = () => {};
    this._start      = opts.start      || noop;
    this._finish     = opts.finish     || noop;
    this._startFlow  = opts.startFlow  || noop;
    this._finishFlow = opts.finishFlow || noop;
    this._startTask  = opts.startTask  || noop;
    this._passTask   = opts.passTask   || noop;
    this._failTask   = opts.failTask   || noop;

    for (const propName in opts) {
      if ([
        'start',
        'finish',
        'startFlow',
        'finishFlow',
        'startTask',
        'passTask',
        'failTask',
      ].includes(propName)) { continue; }
      if (this[propName] !== undefined) {
        throw new Error(`Property name ${propName} is reserved`);
      }
      this[propName] = opts[propName];
    }
  }

  start() {
    this.startedAt = new Date();
    this.flows     = {};
    this._start();
  }

  finish() {
    const flows      = Object.keys(this.flows).map(id => this.flows[id]);
    this.finishedAt  = new Date();
    this.tasks       = flows.reduce((ts, f) => ts.push(...f.tasks) && ts, []);
    this.tasksPassed = this.tasks.filter(t => t.isPass);
    this.tasksFailed = this.tasks.filter(t => !t.isPass);
    this.flowsPassed = flows.filter(f => f.isPass);
    this.flowsFailed = flows.filter(f => !f.isPass);
    this.isPass      = this.flowsFailed.length < 1;
    this.taskTime    = this.tasks.reduce((n, t) => n + t.taskTime, 0);
    this.flowTime    = flows.reduce((n, f) => n + f.flowTime, 0);
    this.totalTime   = this.finishedAt.getTime() - this.startedAt.getTime();
    this._finish();
  }

  startFlow(flowId, description) {
    const flow = {
      tasks    : [],
      startedAt: new Date(),
      description,
    };
    this.flows[flowId] = flow;
    this._startFlow(flow);
  }

  finishFlow(flowId) {
    const flow       = this.flows[flowId];
    flow.finishedAt  = new Date();
    flow.taskTime    = flow.tasks.reduce((n, t) => n + t.taskTime, 0);
    flow.flowTime    = flow.finishedAt.getTime() - flow.startedAt.getTime();
    flow.tasksPassed = flow.tasks.filter(t => t.isPass);
    flow.tasksFailed = flow.tasks.filter(t => !t.isPass);
    flow.isPass      = flow.tasksFailed.length < 1;
    this._finishFlow(flow);
  }

  startTask(flowId, taskIndex, description) {
    assert.ok(
      this.flows[flowId],
      `Started task ${taskIndex} for missing flow at index ${flowId}`
    );

    const flow = this.flows[flowId];
    const task = {
      startedAt: new Date(),
      description,
    };
    flow.tasks[taskIndex] = task;
    this._startTask(task, flow);
  }

  passTask(flowId, taskIndex, result) {
    assert.ok(
      this.flows[flowId],
      `Recorded Pass for task ${taskIndex} which is part of a missing flow at index ${flowId}`
    );
    assert.ok(
      this.flows[flowId].tasks[taskIndex],
      `Recorded Pass for missing task at index ${taskIndex} which is part of a flow at ` +
      `index ${flowId}`
    );

    const flow      = this.flows[flowId];
    const task      = flow.tasks[taskIndex];
    task.finishedAt = new Date();
    task.taskTime   = task.finishedAt.getTime() - task.startedAt.getTime();
    task.isPass     = true;
    task.result     = result;
    this._passTask(task, flow);
  }

  failTask(flowId, taskIndex, error) {
    assert.ok(
      this.flows[flowId],
      `Recorded Fail for task ${taskIndex} which is part of a missing flow at index ${flowId}`
    );
    assert.ok(
      this.flows[flowId].tasks[taskIndex],
      `Recorded Fail for missing task at index ${taskIndex} which is part of a flow at ` +
      `index ${flowId}`
    );

    const flow      = this.flows[flowId];
    const task      = flow.tasks[taskIndex];
    task.finishedAt = new Date();
    task.taskTime   = task.finishedAt.getTime() - task.startedAt.getTime();
    task.isPass     = false;
    task.error      = error;
    this._failTask(task, flow);
  }
}


module.exports = Reporter;
