// eslint-disable-next-line node/no-deprecated-api
const domain       = require('domain');
const relign       = require('relign');
const specReporter = require('../reporter/spec');


class Suite {

  constructor() {
    this.reporter = specReporter;
    this.flows    = [];
  }

  addFlow(flow) {
    for (let i = 1; i < flow.tasks.length; i += 1) {
      flow.tasks[i].previousTask = flow.tasks[i - 1];
    }
    flow.finalTask = flow.tasks[flow.tasks.length - 1];

    this.flows.push(flow);
  }

  setReporter(reporter) {
    this.reporter = reporter;
  }

  async start() {
    this.reporter.start();

    const flows = this.flows.slice(0);

    const getUnblockedFlows = () => flows.filter((flow, i) => {
      const unblocked = !flow.dependencyIds || !flow.dependencyIds.length < 1;
      if (unblocked) { flows.splice(i, 1); }
      return unblocked;
    });

    const runTask = async (flow, task, taskIndex) => {
      this.reporter.startTask(flow.id, taskIndex, task.description);

      let previousResult;
      if (task.previousTask) {
        previousResult = task.previousTask.result;
      } else if (flow.dependencyResults) {
        previousResult = typeof flow.dependencyResults === 'object' ?
          Object.keys(flow.dependencyResults).map(id => flow.dependencyResults[id]) :
          flow.dependencyResults;
      }
      try {
        task.result = await this._executeTask(task.fn, previousResult);
        this.reporter.passTask(flow.id, taskIndex, task.result);
      } catch (err) {
        task.error  = err;
        this.reporter.failTask(flow.id, taskIndex, err);
      }
    };

    const runFlow = async (flow) => {
      this.reporter.startFlow(flow.id, flow.description);

      await relign.seriesMap(flow.tasks, (task, taskIndex) => runTask(flow, task, taskIndex));

      for (const siblingFlow of flows) {
        if (siblingFlow.dependencyIds) {

          if (siblingFlow.dependencyIds === flow.id) {
            siblingFlow.dependencyResults = flow.finalTask.result;
            delete siblingFlow.dependencyIds;
            continue;
          }

          siblingFlow.dependencyResults || (siblingFlow.dependencyResults = {});
          siblingFlow.dependencyResults[flow.id] = flow.finalTask.result;

          siblingFlow.dependencyIds = siblingFlow.dependencyIds.filter(id => flow.id === id);
          if (siblingFlow.dependencyIds.length < 1) {
            delete siblingFlow.dependencyIds;
          }
        }
      }

      this.reporter.finishFlow(flow.id);
    };

    const runFlows = async (flows) => {
      await relign.seriesMap(flows, runFlow);
      const nextFlows = getUnblockedFlows();
      if (nextFlows.length < 1) { return; }
      return runFlows(nextFlows);
    };

    try {
      await runFlows(getUnblockedFlows());
    } catch (err) {
      console.error(err);
    }

    this.reporter.finish();
    return this.reporter.isPass;
  }

  _executeTask(taskFn, previousTaskResult) {
    return new Promise((resolve, reject) => {
      const taskDomain = domain.create();
      taskDomain.on('error', err => reject(err));
      taskDomain.run(() => {
        relign.exec(() => taskFn(previousTaskResult)).then(resolve, reject);
      });
    });
  }
}


module.exports = Suite;
